import json
import os

import pytest

from tktl.commands.login import LogInCommand, SetApiKeyCommand
from tktl.core.exceptions import InvalidInputError


def test_set_api_key(capsys):
    cmd = SetApiKeyCommand()
    cmd.execute(api_key=None)
    out, err = capsys.readouterr()
    assert "API Key cannot be empty.\n" == err

    cmd.execute(api_key="ABC")
    assert os.path.exists(os.path.expanduser("~/.tktl/config.json"))
    with open(os.path.expanduser("~/.tktl/config.json"), "r") as j:
        d = json.load(j)
        assert d["api-key"] == "ABC"


def test_login(capsys, user_password_key):
    cmd = LogInCommand()
    u, p, k = user_password_key
    assert cmd.execute(k) is True
    out, err = capsys.readouterr()
    assert out == "Login successful!\n"
    with pytest.raises(InvalidInputError):
        cmd.execute(None)



from .. import loggers as sdk_logger


class BaseClient(object):
    def __init__(self, api_key, ps_client_name=None, logger=sdk_logger.MuteLogger()):
        """
        Base class. All client classes inherit from it.

        :param str api_key: your API key
        :param str ps_client_name:
        :param sdk_logger.Logger logger:
        """
        self.api_key = api_key
        self.ps_client_name = ps_client_name
        self.logger = logger



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `tktl` package."""
import json
import os

import pytest
from click.testing import CliRunner

from tktl import main


@pytest.fixture
def response():
    """Sample pytest fixture.

    See more at: http://doc.pytest.org/en/latest/fixture.html
    """
    # import requests
    # return requests.get('https://github.com/audreyr/cookiecutter-pypackage')


def test_content(response):
    """Sample pytest test function with the pytest fixture as an argument."""
    # from bs4 import BeautifulSoup
    # assert 'GitHub' in BeautifulSoup(response.content).title.string


def test_command_line_interface():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(main.cli)
    assert result.exit_code == 0
    assert "Show this message and exit." in result.output
    help_result = runner.invoke(main.cli, ["--help"])
    assert help_result.exit_code == 0
    assert "Show this message and exit." in help_result.output


def test_ci_auth():
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(main.login, ["--api-key", os.environ["TEST_API_KEY"]])
    print(result.output)
    assert result.exit_code == 0
    assert result.output == "Login successful!\n"

    result = runner.invoke(main.login, ["--api-key", "FAKEAPIKEY"])
    assert result.exit_code == 0
    assert result.output == "Request failed: Key format is invalid\n"
    with open(os.path.expanduser("~/.tktl/config.json"), "r") as j:
        d = json.load(j)
    assert d["api-key"] == "FAKEAPIKEY"



import click

from tktl.cli.common import ClickGroup
from tktl.commands import deployments as deployments_commands
from tktl.core.config import settings
from tktl.core.loggers import CliLogger
from tktl.core.t import DeploymentStatesT
from tktl.login import validate_decorator


@click.group(
    "deployments",
    help="Manage deployments",
    cls=ClickGroup,
    **settings.HELP_COLORS_DICT
)
def deployments():
    pass


_deployment_shared_options = [
    click.option("-c", "--commit-sha", help="SHA commit"),
    click.option("-b", "--branch-name", help="Branch Name"),
    click.option("-s", "--status-name", help="Status in which deployment is",),
]


def deployment_shared_options(func):
    for option in reversed(_deployment_shared_options):
        func = option(func)
    return func


@deployments.command("get-id", help="Create new deployment")
@deployment_shared_options
@validate_decorator
def get_id(commit_sha: str, branch_name: str, status_name: str, **kwargs):
    logger_ = CliLogger()
    command = deployments_commands.GetDeploymentId()
    result = command.execute(
        commit_sha=commit_sha, branch_name=branch_name, status_name=status_name
    )
    logger_.log(result)


@deployments.command("get-ecr-repo", help="Create new deployment")
@deployment_shared_options
@validate_decorator
def get_ecr_repo(commit_sha: str, branch_name: str, status_name: str, **kwargs):
    logger_ = CliLogger()
    command = deployments_commands.GetDeploymentEcr()
    result = command.execute(
        commit_sha=commit_sha, branch_name=branch_name, status_name=status_name
    )
    logger_.log(result)


@deployments.command("patch-status", help="Create new deployment")
@click.option(
    "-id", "--deployment-id", help="Deployment id", required=True,
)
@click.option(
    "-s",
    "--status-name",
    help="Patch the status of a deployment",
    type=click.Choice(DeploymentStatesT.list()),
)
@validate_decorator
def patch_status(deployment_id: str, status_name: str, **kwargs):
    logger_ = CliLogger()
    command = deployments_commands.PatchDeploymentStatus()
    result = command.execute(deployment_id=deployment_id, status_name=status_name)
    logger_.log(result)



import json
from functools import wraps

import requests

from tktl.core.config import settings
from tktl.core.exceptions import InvalidInputError, AuthenticationError
from tktl.core.loggers import CliLogger
from tktl.core.managers.auth import AuthConfigManager
from tktl.core.strings import CLIStrings, HeaderStrings

UNAUTHORIZED_EXTENDED_INFO = (
    "\n\nNote: Please keep in mind that currently you can login only with the email and "
    "password from your Paperspace account. If you're using AD, SAML or GitHub "
    "credentials, please log into the Paperspace Console and create an API key for use with "
    "the CLI client. For more information, please refer to the CLI client documentation."
)

logger = CliLogger()


def is_error_or_missing_keys_print(res, *required_keys):
    if "error" in res:
        if "message" in res:
            print(res["message"])
            return True
        if "message" in res["error"]:
            error_message = res["error"]["message"] + UNAUTHORIZED_EXTENDED_INFO
            print(error_message)
            return True
        print(json.dumps(res, indent=2, sort_keys=True))
        return True
    elif not all(key in res for key in required_keys):
        if "message" in res:
            print(res["message"])
            return True
        print(json.dumps(res, indent=2, sort_keys=True))
        return True
    return False


def login(api_key=None):
    """

    Parameters
    ----------
    api_key

    Returns
    -------

    """

    if not api_key:
        raise InvalidInputError(
            CLIStrings.INVALID_INPUT + "api_key must be provided"
        )

    AuthConfigManager.set_api_key(api_key)
    return True


def set_api_key(api_key):
    AuthConfigManager.set_api_key(api_key=api_key)
    path = (
        f"{AuthConfigManager.get_tktl_config_path()}/"
        f"{AuthConfigManager.SETTINGS.CONFIG_FILE_NAME}"
    )
    logger.log(f"Successfully added your API Key to {path} You're ready to go!")
    return True


def logout():
    AuthConfigManager.clear_tktl_config()
    return True


def validate_key(suppress=False):
    key = AuthConfigManager.get_api_key()
    if not key:
        logger.error(CLIStrings.AUTH_ERROR_MISSING_KEY)
        raise AuthenticationError(CLIStrings.AUTH_ERROR_MISSING_KEY)

    r = requests.get(
        f"{settings.TAKTILE_API_HOST}{settings.API_V1_STR}/users/me",
        headers={"Accept": HeaderStrings.APPLICATION_JSON, "X-Api-Key": key},
    )
    try:
        r.raise_for_status()
    except requests.exceptions.RequestException as e:
        logger.error(f'Request failed: {r.json()["detail"]}')
        return False

    if not suppress:
        logger.log("Login successful!", color="green")
    return True


def validate_decorator(func):
    @wraps(func)
    def wrapped_validation(*args, **kwargs):
        validate_key(suppress=True)
        return func(*args, **kwargs)

    return wrapped_validation



import getpass

import click

from tktl.commands import login as login_commands
from tktl.core.loggers import CliLogger

logger = CliLogger()


@click.command("login", help="Perform authenticated requests against the taktile api")
@click.option("-k", "--api-key", "api_key", help="Your api key")
def login(api_key: str):
    command = login_commands.LogInCommand()
    command.execute(api_key=api_key)


@click.command("logout", help="Log out / remove apiKey from config file")
def logout():
    command = login_commands.LogOutCommand()
    command.execute()


@click.command("api-key", help="Save your api key")
@click.argument(
    "api_key", required=False,
)
def save_api_key(api_key):
    if not api_key:
        api_key = getpass.getpass("Enter your API Key: ")

    command = login_commands.SetApiKeyCommand()
    api_key = api_key.strip()
    command.execute(api_key)



from typing import Optional

from pydantic import BaseModel, UUID4

from tktl.core.t import DeploymentStatesT


class DeploymentBase(BaseModel):
    id: Optional[UUID4] = None
    git_branch: str = None
    git_hash: str = None
    requester_id: UUID4 = None
    repository_id: UUID4 = None
    repo_ref_id: int
    status_id: UUID4 = None
    ecr_repo_url: str


# Properties to receive on item update
class DeploymentUpdate(DeploymentBase):
    status_name: DeploymentStatesT



import numpy as np
from sklearn.metrics import roc_auc_score
from src.endpoints import binary, X_test, y_test


def test_accuracy(json_metadata):
    lower_bound = 0.85
    pred = binary(X_test)
    auc = roc_auc_score(y_test, pred)
    assert auc > lower_bound

    # add metadata
    json_metadata["section"] = "Accuracy"
    json_metadata["pass_message"] = f"AUC exceeds lower bound of {lower_bound}"
    json_metadata["fail_message"] = f"AUC is below lower bound of {lower_bound}"


def test_mean(json_metadata):
    pred = binary(X_test)
    mean_pred = np.mean(pred)
    assert 0.3 < mean_pred < 0.4

    # add metadata
    json_metadata["section"] = "Plausibility"
    json_metadata["pass_message"] = f"Mean prediction is plausible ({mean_pred:.2f})"
    json_metadata["fail_message"] = f"Mean prediction is implausible ({mean_pred:.2f})"



from tktl import __version__
from tktl.commands import CommandBase
from tktl.login import login, logout, validate_key, set_api_key


class LogInCommand(CommandBase):
    def execute(self, api_key=None):
        if login(api_key):
            return validate_key()
        return False


class LogOutCommand(CommandBase):
    def execute(self):
        logout()


class ShowVersionCommand(CommandBase):
    def execute(self):
        self.logger.log(__version__)


class SetApiKeyCommand(CommandBase):
    def execute(self, api_key):
        if not api_key:
            self.logger.error("API Key cannot be empty.")
            return
        set_api_key(api_key=api_key)



import logging

import click

from tktl.commands.init import init_project

_logger = logging.getLogger("root")


@click.command()
@click.option("--name", help="Name of the project", default="tktl-serving")
@click.option(
    "--path", help="Directory where the new project will be created", default="."
)
def init(path: str, name: str) -> None:
    """Creates a new project with the necessary scaffolding, as well as the supporting
        files needed. The directory structure of a new project , and the files within it
        will look like this::

            regression.py         # Regression starter file
            classification.py     # Classification starter file
            racket.yaml           # Main project config file
            .gitignore            # Just in case
            Dockerfile.tfserving  # TFS' Dockerfile]
            Dockerfile.racket     # racket's Dockerfile. To be implemented
            docker-compose.yaml   # Docker-compose to start up TFS
            serialized/           # Where the serialized models will be stored

        """
    init_project(path=path, name=name)



import os

from tktl.core.loggers import CliLogger
from tktl.core.managers.project import ProjectManager

logger = CliLogger()


def init_project(path: str, name: str):
    ProjectManager.init_project(path, name)
    logger.log(
        f"Project scaffolding created successfully "
        f"at {os.path.abspath(path)}/{name}",
        color="green",
    )



import click

from tktl.commands import login as login_commands


@click.command("version", help="Show the version and exit")
def get_version():
    command = login_commands.ShowVersionCommand()
    command.execute()



import abc

from click import secho

from tktl.core.config import settings


class Logger(abc.ABC):
    @abc.abstractmethod
    def log(self, msg, *args, **kwargs):
        pass

    @abc.abstractmethod
    def warning(self, msg, *args, **kwargs):
        pass

    @abc.abstractmethod
    def error(self, msg, *args, **kwargs):
        pass

    def debug(self, msg, *args, **kwargs):
        pass


class MuteLogger(Logger):
    def log(self, msg, *args, **kwargs):
        pass

    def warning(self, msg, *args, **kwargs):
        pass

    def error(self, msg, *args, **kwargs):
        pass


class CliLogger(Logger):
    @staticmethod
    def _log(message, color=None, err=False):
        message = str(message)
        color = color if settings.USE_CONSOLE_COLORS else None
        secho(message, fg=color, err=err)

    def log(self, message, color=None, err=False):
        self._log(message, color=color, err=err)

    def error(self, message, *args, **kwargs):
        color = "red" if settings.USE_CONSOLE_COLORS else None
        self._log(message, color=color, err=True)

    def warning(self, message, *args, **kwargs):
        color = "yellow" if settings.USE_CONSOLE_COLORS else None
        self._log(message, color=color)

    def debug(self, message, *args, **kwargs):
        if settings.DEBUG:
            self._log("DEBUG: {}".format(message))



import logging

import click

from tktl.commands.validate import validate_project_config

_logger = logging.getLogger("root")


@click.command()
@click.option(
    "--path", "-p", help="Directory where the new project will be created", default="."
)
def validate(path) -> None:
    """Validates a new project for the necessary scaffolding, as well as the supporting
        files needed. The directory structure of a new project, and the files within it
        should look like this::

            regression.py         # Regression starter file
            classification.py     # Classification starter file
            tktl.yaml             # Main project config file
            .gitignore            # Just in case
            Dockerfile.tfserving  # TFS' Dockerfile]
            Dockerfile.racket     # racket's Dockerfile. To be implemented
            docker-compose.yaml   # Docker-compose to start up TFS

        """

    validate_project_config(path=path)



import numpy as np
import pandas as pd
import pytest

from tktl import Tktl
from tktl.core.exceptions.exceptions import ValidationException


@pytest.mark.parametrize(
    "X,y",
    [
        (pd.DataFrame({"a": [1, 2, 3]}), pd.Series([4, 5, 5])),  # pandas
        ({"a": [1, 2, 3]}, [4, 5, 5]),  # base
    ],
)
def test_creation(X, y):
    tktl = Tktl()

    @tktl.endpoint(X=X, y=y, kind="regression")
    def predict(X):
        return [0] * len(X)

    endpoint = tktl.endpoints[0]
    assert type(endpoint.X) == pd.DataFrame
    assert type(endpoint.y) == pd.Series
    assert endpoint.y.name == "Outcome"
    assert endpoint.func(pd.DataFrame(X)) == [0, 0, 0]


def test_validate_func():
    with pytest.raises(ValidationException):
        tktl = Tktl()
        X = {"a": [1, 2, 3]}
        y = [4, 5, 5]

        @tktl.endpoint(X=X, y=y)
        def predict(X):
            return [0] * (len(X) - 1)


def test_input_dimensions():
    with pytest.raises(ValidationException):
        tktl = Tktl()
        X = {"a": [1, 2, 3]}
        y = [4, 5]

        @tktl.endpoint(X=X, y=y)
        def predict(X):
            return [0] * len(X)


def test_func_dimensions():
    with pytest.raises(ValidationException):
        tktl = Tktl()
        X = {"a": [1, 2, 3]}
        y = [4, 5, 5]

        @tktl.endpoint(X=X, y=y)
        def predict(X):
            return [0] * (len(X) - 1)


def test_unknown_kind():
    with pytest.raises(NotImplementedError):
        tktl = Tktl()
        X = {"a": [1, 2, 3]}
        y = [4, 5, 5]

        @tktl.endpoint(X=X, y=y, kind="unknown kind")
        def predict(X):
            return [0] * len(X)


def test_missing():
    tktl = Tktl()
    X = {"a": [1, 2, 3]}
    y = pd.Series([0.1, 0.2, None])

    with pytest.warns(UserWarning):

        @tktl.endpoint(X=X, y=y, kind="regression")
        def predict(X):
            return [0] * len(X)


def test_numpy_arrays():
    tktl = Tktl()
    X = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    y = np.array([0.1, 0.2, 0.3])

    @tktl.endpoint(X=X, y=y, kind="regression")
    def predict(X):
        return np.array([0] * len(X))

    endpoint = tktl.endpoints[0]
    assert type(endpoint.X) == pd.DataFrame
    assert type(endpoint.y) == pd.Series
    assert endpoint.y.name == "Outcome"
    assert np.allclose(endpoint.func(pd.DataFrame(X)), np.array([0, 0, 0]))



from tktl.core import ExtendedEnum


class UserRepoConfigFileNameT(str, ExtendedEnum):
    YAML = "tktl.yaml"
    YML = "tktl.yml"


class UserProjectFileT(str, ExtendedEnum):
    FILE = "file"
    DIRECTORY = "dir"


class RequiredUserProjectPathsT(str, ExtendedEnum):
    SRC = "src"
    TKTL_YML = "tktl.yml"
    TKTL_YAML = "tktl.yaml"
    DOCKERFILE = "Dockerfile"
    REQS = "requirements.txt"

    @classmethod
    def strictly_required_files(cls):
        return {cls.DOCKERFILE.value, cls.REQS.value}

    @classmethod
    def strictly_required_dirs(cls):
        return {cls.SRC.value}


class DeploymentStatesT(str, ExtendedEnum):
    TEST_FAILED = "test_failed"
    IMAGE_BUILD_FAILED = "image_build_failed"
    DEPLOYMENT_FAILED = "deployment_failed"
    CANCELED = "canceled"
    SUCCEEDED = "succeeded"
    TEST_RUNNING = "test_running"
    IMAGE_BUILDING = "image_building"
    DEPLOYING = "deploying"
    MARKED_FOR_DELETION = "marked_for_deletion"
    DELETION_FAILED = "deletion_failed"


class ProjectAssetT(str, ExtendedEnum):
    DATA = "data"
    MODEL = "model"


class ProjectAssetSourceT(str, ExtendedEnum):
    S3 = "s3"
    LOCAL = "local"
    LFS = "lfs"


class ServiceT(str, ExtendedEnum):
    REST = "rest"
    GRPC = "grpc"


class ServiceComputeT(str, ExtendedEnum):
    GPU = "gpu"
    CPU = "cpu"



import copy
from typing import Type

import requests
from pydantic import BaseModel

from tktl.core import utils, loggers as sdk_logger
from tktl.core.exceptions.exceptions import APIClientException
from tktl.core.managers.auth import AuthConfigManager


class API(object):
    DEFAULT_HEADERS = {
        "Accept": "application/json",
    }

    def __init__(self, api_url, headers=None, logger=sdk_logger.MuteLogger()):
        """

        :param str api_url: url you want to connect
        :param dict headers: headers
        :param str api_key: your API key
        :param str ps_client_name: Client name
        :param sdk_logger.Logger logger:
        """
        self.api_url = api_url
        headers = headers or self.DEFAULT_HEADERS
        self.headers = headers.copy()
        self.logger = logger
        self.api_key = AuthConfigManager.get_api_key()
        self.DEFAULT_HEADERS.update({"X-Api-Key": self.api_key})

    def get_path(self, url=None):
        if not url:
            return self.api_url
        full_path = utils.concatenate_urls(self.api_url, url)
        return full_path

    def post(self, url=None, json=None, params=None, files=None, data=None):
        path = self.get_path(url)
        headers = copy.deepcopy(self.DEFAULT_HEADERS)
        if data:
            headers["Content-Type"] = data.content_type

        self.logger.debug(
            "POST request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}\n\tfiles: {}\n\tdata: {}".format(
                path, headers, json, params, files, data
            )
        )
        response = requests.post(
            path, json=json, params=params, headers=headers, files=files, data=data
        )
        self.logger.debug("Response status code: {}".format(response.status_code))
        self.logger.debug("Response content: {}".format(response.content))
        return response

    def put(self, url=None, json=None, params=None, data=None):
        path = self.get_path(url)
        self.logger.debug(
            "PUT request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                path, self.headers, json, params
            )
        )
        response = requests.put(
            path, json=json, params=params, headers=self.DEFAULT_HEADERS, data=data
        )
        self.logger.debug("Response status code: {}".format(response.status_code))
        self.logger.debug("Response content: {}".format(response.content))
        return response

    def patch(self, url=None, json=None, params=None, data=None):
        path = self.get_path(url)
        self.logger.debug(
            "PATCH request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                path, self.headers, json, params
            )
        )
        response = requests.patch(
            path, json=json, params=params, headers=self.DEFAULT_HEADERS, data=data
        )
        self.logger.debug("Response status code: {}".format(response.status_code))
        self.logger.debug("Response content: {}".format(response.content))
        return response

    def get(self, url=None, json=None, params=None):
        path = self.get_path(url)

        self.logger.debug(
            "GET request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                path, self.headers, json, params
            )
        )
        response = requests.get(
            path, params=params, headers=self.DEFAULT_HEADERS, json=json
        )
        self.logger.debug("Response status code: {}".format(response.status_code))
        self.logger.debug("Response content: {}".format(response.content))
        return response

    def delete(self, url=None, json=None, params=None):
        path = self.get_path(url)
        response = requests.delete(path, params=params, headers=self.headers, json=json)
        self.logger.debug(
            "DELETE request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                response.url, self.headers, json, params
            )
        )
        self.logger.debug("Response status code: {}".format(response.status_code))
        self.logger.debug("Response content: {}".format(response.content))
        return response


class TaktileResponse(object):
    def __init__(self, body, code, headers, data):
        self.body = body
        self.code = code
        self.headers = headers
        self.data = data

    @property
    def ok(self):
        return 200 <= self.code < 400

    @classmethod
    def interpret_response(cls, response: requests.Response, model: Type[BaseModel]):
        """
        Parameters
        ----------
        response

        model
        """

        try:
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            raise APIClientException(status_code=e.response.status_code, detail=repr(e))
        return model(**response.json())



import pytest

from tktl.commands.deployments import (
    GetDeployment,
    GetDeploymentEcr,
    GetDeploymentId,
    PatchDeploymentStatus,
)
from tktl.core.exceptions.exceptions import APIClientException


def test_get_deployment_commands():
    cmd = GetDeployment()
    with pytest.raises(APIClientException) as e:
        cmd.execute("fake", "branch", "name")

    cmd = GetDeploymentEcr()
    with pytest.raises(APIClientException) as e:
        cmd.execute("fake", "branch", "name")

    cmd = GetDeploymentId()
    with pytest.raises(APIClientException) as e:
        cmd.execute("fake", "branch", "name")

    cmd = PatchDeploymentStatus()
    with pytest.raises(APIClientException) as e:
        cmd.execute("fake", "status")

    # print(response)
    # repr(response)
    # print(response.dict())
    # assert response.status_code == 423

    #
    # out, err = capsys.readouterr()
    # assert 'API Key cannot be empty.\n' == err
    #
    # cmd.execute(api_key='ABC')
    # assert os.path.exists(os.path.expanduser('~/.tktl/config.json'))
    # with open(os.path.expanduser('~/.tktl/config.json'), 'r') as j:
    #     d = json.load(j)
    #     assert d['api-key'] == 'ABC'


# def test_login(capsys, user_password_key):
#     cmd = LogInCommand()
#     with pytest.raises(InvalidInputError):
#         cmd.execute(None, None, None)
#
#     with pytest.raises(InvalidInputError):
#         cmd.execute(None, 'me', None)
#
#     u, p, k = user_password_key
#
#     assert cmd.execute(u, p, None) is True
#     out, err = capsys.readouterr()
#     assert out == 'Login successful!\n'
#     assert cmd.execute(None, None, k) is True
#     out, err = capsys.readouterr()
#     assert out == 'Login successful!\n'
#
#     assert cmd.execute(u, 'whatever', None) is False
#     out, err = capsys.readouterr()
#     assert err == 'Request failed: Incorrect username or password\n'
#



from dataclasses import dataclass
import functools
import warnings
from typing import Callable

import pandas as pd

from tktl.core.exceptions import exceptions


@dataclass
class Endpoint:
    """Container for endpoint components"""

    func: Callable
    kind: str
    X: pd.DataFrame
    y: pd.Series


class Tktl:
    def __init__(self):
        self.endpoints = []

    # This is the user-facing decorator for function registration
    def endpoint(
        self,
        func: Callable = None,
        X: pd.DataFrame = None,
        y: pd.Series = None,
        kind: str = "regression",
    ):
        """Register function as a Taktile endpoint

        Parameters
        ----------
        func : Callable, optional
            Function that describes the desired operation, by default None
        X : pd.DataFrame, optional
            Reference input dataset for testing func, by default None
        y : pd.Series, optional
            Reference output for evaluating func, by default None
        kind : str, optional
            Specification of endpoint type ("regression", "binary"), by default "regression"

        Returns
        -------
        Callable
            Wrapped function
        """
        if func is None:
            return functools.partial(self.endpoint, kind=kind, X=X, y=y)

        X, y = self._convert_xy(X, y, kind=kind)
        self._validate_func(func, X, y)
        endpoint = Endpoint(func, kind, X=X, y=y)
        self.endpoints.append(endpoint)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            pred = func(*args, **kwargs)
            return pred

        return wrapper

    def _convert_xy(self, X, y, kind):
        # Convert X to DataFrame
        try:
            X = pd.DataFrame(X)
        except ValueError:
            raise exceptions.ConversionException(
                "Could not convert X to pandas dataframe"
            )

        if len(X) > 1e6:
            warnings.warn(
                f"X is very large (N={len(X)}). "
                f"Please consider using a smaller reference dataset."
            )

        # Convert y to Series
        try:
            y = pd.Series(y)
            y.name = y.name or "Outcome"

            if kind == "regression":
                y = y.astype(float)
            elif kind == "binary":
                y = y.astype(bool)
            else:
                raise NotImplementedError(f"Unknown endpoint kind: {kind}")

        except ValueError:
            raise exceptions.ConversionException("Could not convert y to pandas series")

        # Drop observations with missing outcomes
        id_missing = y.isna()
        n_missing = id_missing.sum()
        if id_missing.sum() > 0:
            warnings.warn(f"y contains {n_missing} missing values that will be dropped")
            X = X.loc[~id_missing]
            y = y.loc[~id_missing]

        if y.shape != (len(X),):
            raise exceptions.ValidationException("y is inconsistent with input data X")

        return X, y

    def _validate_func(self, func, X, y):
        pred = func(X)
        try:
            pred = list(pred)
        except ValueError:
            raise exceptions.ConversionException(
                "Could not convert function output to list"
            )

        if len(pred) != len(X):
            raise exceptions.ValidationException(
                "Predictions are inconsistent with input data X"
            )



from pydantic import ValidationError

from tktl.core.exceptions.exceptions import (
    NoContentsFoundException,
    UserRepoValidationException,
)
from tktl.core.loggers import CliLogger
from tktl.core.managers.project import ProjectManager
from tktl.core.schemas.project import ProjectValidationOutput
from tktl.core.validation.outputs import (
    ProjectValidationFailure,
    ConfigFileValidationFailure,
)

logger = CliLogger()


def validate_project_config(path: str):
    try:
        ProjectManager.validate_project_config(path)
    except ValidationError as e:
        validation_output = ProjectValidationOutput(
            title=ConfigFileValidationFailure.title,
            summary=ConfigFileValidationFailure.summary,
            text=ConfigFileValidationFailure.format_step_results(validation_errors=e),
        )
        log_failure(validation_output)
        return
    except (NoContentsFoundException, UserRepoValidationException) as e:
        validation_output = ProjectValidationOutput(
            title=ProjectValidationFailure.title,
            summary=ProjectValidationFailure.summary,
            text=ProjectValidationFailure.format_step_results(validation_errors=e),
        )
        log_failure(validation_output)
        return
    logger.log(f"Project scaffolding is valid!", color="green")


def log_failure(validation_output: ProjectValidationOutput):
    logger.log(
        f"Project scaffolding is invalid: {validation_output.title}", color="red"
    )
    logger.log(validation_output.summary, color="red", err=True)
    logger.log(validation_output.text, color="red", err=True)



import functools
import re

import click
import termcolor
from click_didyoumean import DYMMixin
from click_help_colors import HelpColorsGroup

from tktl.core.config import settings

OPTIONS_FILE_OPTION_NAME = "optionsFile"
OPTIONS_FILE_PARAMETER_NAME = "options_file"
OPTIONS_DUMP_FILE_OPTION_NAME = "createOptionsFile"


class ClickGroup(DYMMixin, HelpColorsGroup):
    pass


def deprecated(msg):
    deprecated_invoke_notice = (
        msg
        + """\nFor more information, please see:

https://docs.taktile.com
If you depend on functionality not listed there, please file an issue."""
    )

    def new_invoke(self, ctx):
        click.echo(click.style(deprecated_invoke_notice, fg="red"), err=True)
        super(type(self), self).invoke(ctx)

    def decorator(f):
        f.invoke = functools.partial(new_invoke, f)

    return decorator


class ColorExtrasInCommandHelpMixin(object):  # noqa
    def get_help_record(self, *args, **kwargs):
        rv = super(ColorExtrasInCommandHelpMixin, self).get_help_record(*args, **kwargs)
        if not settings.USE_CONSOLE_COLORS:
            return rv

        try:
            help_str = rv[1]
        except (IndexError, TypeError):
            return rv

        if help_str:  # noqa
            help_str = self._color_extras(help_str)
            rv = rv[0], help_str
        return rv

    def _color_extras(self, s):
        pattern = re.compile(r"^.*(\[.*\])$")
        found = re.findall(pattern, s)
        if found:
            extras_str = found[-1]
            coloured_extras_str = self._color_str(extras_str)
            s = s.replace(extras_str, coloured_extras_str)

        return s

    def _color_str(self, s):
        s = termcolor.colored(s, settings.HELP_HEADERS_COLOR)
        return s



from tktl.commands import BaseDeploymentApiCommand
from tktl.core.clients.http_client import TaktileResponse
from tktl.core.config import settings
from tktl.core.schemas.deployment import DeploymentBase


class GetDeployment(BaseDeploymentApiCommand):
    def execute(self, commit_sha: str, branch_name: str, status_name: str):
        params = {}
        if branch_name:
            params.update({"branch_name": branch_name})
        if status_name:
            params.update({"status": status_name})
        response = self.client.get(
            f"{settings.API_V1_STR}/deployments/{commit_sha}", params=params
        )
        return TaktileResponse.interpret_response(
            response=response, model=DeploymentBase
        )


class GetDeploymentId(GetDeployment):
    def execute(self, commit_sha: str, branch_name: str, status_name: str):
        return (
            super()
            .execute(
                commit_sha=commit_sha, branch_name=branch_name, status_name=status_name
            )
            .id
        )


class GetDeploymentEcr(GetDeployment):
    def execute(self, commit_sha: str, branch_name: str, status_name: str):
        return (
            super()
            .execute(
                commit_sha=commit_sha, branch_name=branch_name, status_name=status_name
            )
            .ecr_repo_url
        )


class PatchDeploymentStatus(GetDeployment):
    def execute(self, deployment_id: str, status_name: str, **kwargs):
        response = self.client.patch(
            f"{settings.API_V1_STR}/deployments/{deployment_id}/status",
            params={"updated_status": status_name},
        )
        return TaktileResponse.interpret_response(response, DeploymentBase)


class ListDeploymentsCommand(BaseDeploymentApiCommand):
    WAITING_FOR_RESPONSE_MESSAGE = "Waiting for data..."

    def execute(self, **kwargs):
        # TODO: implement list deployments and show as table

        # with halo.Halo(text=self.WAITING_FOR_RESPONSE_MESSAGE, spinner="dots"):
        #     instances = self._get_instances(**kwargs)

        raise NotImplemented


class GetDeploymentDetails(BaseDeploymentApiCommand):
    def _get_table_data(self, instance):
        pass

    def execute(self, id_):
        # TODO: get deployment details
        raise NotImplemented


class GetDeploymentMetricsCommand(BaseDeploymentApiCommand):
    def execute(
        self, deployment_id, start, end, interval, built_in_metrics, *args, **kwargs
    ):
        # TODO: stream metrics
        raise NotImplemented


class StreamDeploymentMetricsCommand(BaseDeploymentApiCommand):
    def execute(self, **kwargs):
        # TODO: stream metrics
        raise NotImplemented


class DeploymentLogsCommand(BaseDeploymentApiCommand):
    def execute(self, **kwargs):
        # TODO: get logs
        raise NotImplemented

    def _get_log_row_string(self, id, log):
        raise NotImplemented

    def _make_table(self, logs, id):
        raise NotImplemented



import joblib
import pandas as pd
from tktl import Tktl

# Instantiate Taktile client
tktl = Tktl()

# Load reference data
data_test = pd.read_parquet("assets/test.pqt")
label = "survived"
X_test = data_test.drop(columns=label)
y_test = data_test[label]

# Load model
model = joblib.load("assets/model.pkl")


# Define predict functions
@tktl.endpoint(kind="binary", X=X_test, y=y_test)
def binary(X):
    pred = model.predict_proba(X)[:, 1]
    return pred
